SET statement_timeout = 0;
SET client_encoding = 'UTF8';
SET standard_conforming_strings = on;
SET check_function_bodies = false;
SET client_min_messages = warning;

SET default_tablespace = '';

SET default_with_oids = false;

create schema if not exists i3dh;

alter schema l3d owner to i3dhub;

alter role i3dhub set search_path to l3d, i3dh, extensions;

-- cleaup

drop index if exists resource_parameters_res_id_idx;
drop table if exists i3dh.resource_parameters;

drop index if exists resource_endpoints_res_id_idx;
drop table if exists i3dh.resource_endpoints;

drop index if exists resource_res_id_idx;
drop index if exists resource_id_idx;
drop index if exists resource_date_create_id_idx;
drop table if exists i3dh.resource;

-- create

create table i3dh.resource
(
    id                 bigserial   not null,
    res_id             bpchar(40)  not null,           -- could be bytea but would make manual queries annoying
    "type"             text        not null,           -- TODO: extra table
    state              int4        not null,
    date_create        timestamptz not null default now(),
    date_last_read     timestamptz not null default now(),
    "descriptor"       json        not null default '{}'::json,
    endpoint_target_id text        not null default '',
    revision           int4        not null default 0, -- set default to 0
    constraint resource_pkey primary key (id)
);

alter table i3dh.resource
    owner to i3dhub;

create unique index resource_res_id_idx on i3dh.resource using btree (res_id);
create unique index resource_id_idx on i3dh.resource using btree (id);
create unique index resource_date_create_id_idx on i3dh.resource using btree (date_create desc, id desc);


create table i3dh.resource_endpoints
(
    id        bigserial not null,
    res_id    bigint    not null,
    name      text      not null, -- TODO: extra table
    port      integer   not null,
    base_path text      not null,
    protocol  text      not null,
    constraint resource_endpoints_pkey primary key (id),
    constraint resource_endpoints_fkey foreign key (res_id)
        references i3dh.resource (id)
);

alter table i3dh.resource_endpoints
    owner to i3dhub;

create index resource_endpoints_res_id_idx on i3dh.resource_endpoints using btree (res_id);


create table i3dh.resource_parameters
(
    id         bigserial not null,
    res_id     bigint    not null,
    parameters json      not null,
    constraint resource_parameters_pkey primary key (id),
    constraint resource_parameters_fkey foreign key (res_id)
        references i3dh.resource (id)
);

alter table i3dh.resource_parameters
    owner to i3dhub;

create unique index resource_parameters_res_id_idx on i3dh.resource_parameters using btree (res_id);

create table if not exists i3dh.resource_children
(
    parent_id bigint not null,
    child_id  bigint not null,

    foreign key (parent_id) references i3dh.resource (id)
        on update cascade
        on delete cascade,
    foreign key (child_id) references i3dh.resource (id)
        on update cascade
        on delete cascade,
    primary key (parent_id, child_id)
);

alter table i3dh.resource_children
    owner to i3dhub;
