------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------
--
-- This file contains all necesary stored procedures for copying locally a remote cache's data into database 
--
--
------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------

-- search for extensions in l3d schema
SET search_path = l3d, extensions;

drop function if exists l3d.export_cache(varchar, varchar, boolean, integer);


CREATE OR REPLACE FUNCTION l3d.export_cache(p_uri varchar, p_query varchar, result_as_string boolean, p_max_row_count integer default 1000)
 RETURNS TABLE(query_result character varying)
 LANGUAGE plpgsql
AS $function$
declare
	--- select * from l3d.export_cache('/201af4c330f3ef65-67cd02ca-43a1-ade3/0/5910/data/export','{"cache":{},"nodes" :{}, "aux_nodes":{}}', false)
	--- select * from l3d.export_cache('/201af4c330f3ef65-67cd02ca-43a1-ade3/0/3377/data/export','{"nodes" :{}}', false)
		v_cache_id bigint;
begin
		raise notice 'l3d.export_cache(''%'',''%'')', p_uri, p_query ;
	

		if result_as_string then

			return query select ('[' || coalesce( array_to_string( array_agg(r.query_result), ','), '') || ']')::varchar from l3d.export_cache(p_uri, p_query, p_max_row_count) r;

		else

			return query
			select r.query_result
			from l3d.export_cache(p_uri, p_query, p_max_row_count) r;

	 	end if;

end;
$function$;

alter function l3d.export_cache(varchar, varchar, boolean, integer)  OWNER TO i3dhub;

drop function if exists l3d.export_cache(varchar, varchar, integer);


CREATE OR REPLACE FUNCTION l3d.export_cache(p_uri varchar, p_query varchar, p_max_row_count integer default null)
 RETURNS TABLE(query_result varchar)
 LANGUAGE plpgsql
AS $function$
declare
	-- select * from l3d.export_cache('/201af4c330f3ef65-67cd02ca-43a1-ade3/0/3377/data/export','{"cache":{},"nodes" :{}, "aux_nodes":{}}')	
	-- select * from l3d.export_cache('/201af4c330f3ef65-67cd02ca-43a1-ade3/0/3377/data/export','{"nodes" :{}}')	
	-- select * from l3d.export_cache('/3b3596965903685a-20d69378-4e5a-bf8e/0/472/data/export','{"cache":{},"nodes" :{}, "aux_nodes":{}, "shapes" :{}}')	
	-- select * from l3d.export_cache('/3b3596965903685a-20d69378-4e5a-bf8e/0/472/data/export','{}')	

		v_caches_ids integer[];
begin
		-- -------------------------------------------------------------------------------------------------------------------------------------------------------------
		-- cache on the uri
		if p_uri ~ '[a-fA-F0-9]{40}' then
		
			select array_agg( id )::integer[] into v_caches_ids
			from
			(
				select max(c.id) as id
				from
				(
	
					select 	uuid_v3, 
							uuid_v2[1] || '-' || uuid_v2[2] || '-' || uuid_v2[3] || '-' || uuid_v2[4] as uuid_v2 , 
							( 'x' || uuid_v2[5] )::bit(16)::int as id0 ,
							( 'x' || uuid_v2[6] )::bit(16)::int as id1  
					from
					(
						select  p_uri as uuid_v3 , regexp_matches( p_uri, '([a-fA-F0-9]{16})([a-fA-F0-9]{8})([a-fA-F0-9]{4})([a-fA-F0-9]{4})([a-fA-F0-9]{4})([a-fA-F0-9]{4})' ) uuid_v2 
					) a
				 ) a , l3d.cache c
				where 
					( c.uuid = uuid_v3 or ( c.uuid = uuid_v2 and c.ext_cache_id0 = id0 and c.ext_cache_id1 = id1 ) ) and c.active 
			) a;						
		
		elsif p_uri ~ '[a-fA-F0-9-]{35,60}/[0-9]+/[0-9]+'   then

			select array_agg( cac.id )::integer[]
			into v_caches_ids
			from
			(
				select id[1] as uuid, id[2]::integer as ext_cache_id0, id[3]::integer as ext_cache_id1 
				from ( select regexp_matches( p_uri , '([a-fA-F0-9-]{35,60})/([0-9]+)/([0-9]+)' ) as id ) a  						
			) a , l3d."cache" cac
			where
				cac.uuid  = a.uuid and cac.ext_cache_id0 = a.ext_cache_id0 and cac.ext_cache_id1  = a.ext_cache_id1  and cac.active ;

--			raise notice 'caches %', v_caches_ids;

		else

			if p_query not ilike '%caches%:%{%}%' then
				raise exception 'caches are missing on the query %', p_query;
			end if;

			select array_agg(cac.id)::integer[] 
			into v_caches_ids
			from
			(
				select string_to_array( substring(c, '"([^"]*)"') , '/')as cache_uri
				from
				(
					select 	( regexp_matches (c, '"[a-fA-F0-9-]{35,60}/[0-9]+/[0-9]+"[\s]*:[\s]*[^"]*', 'g')  )[1] as c
					from
					(
						select substring(p_query, '"?caches"?[\s]*:[\s]*[{][\s]*([^}]*)') as c
					) a
				) a
			) a , l3d."cache" cac
			where
				cac.uuid  = a.cache_uri[1] and
				cac.ext_cache_id0 = a.cache_uri[2] ::integer and
				cac.ext_cache_id1  = a.cache_uri[3] ::integer and
				cac.active ;

--			raise notice 'caches %', v_caches_ids;
		
		end if;

		return query select 
			( '{"cache":{' || array_to_string( array_agg( a.data_row )  , ',' )|| ' }}' ) ::varchar
			from l3d.export_cache(v_caches_ids, p_query, p_max_row_count) a , l3d."cache" c 
				where c.id  = a.cache_id
			group by c.id
			;
	 	-- -------------------------------------------------------------------------------------------------------------------------------------------------------------
end;		
$function$;


alter function l3d.export_cache(varchar, varchar, integer)  OWNER TO i3dhub;


drop function if exists l3d.export_cache(integer[], varchar, integer);

CREATE OR REPLACE FUNCTION l3d.export_cache(p_caches_ids integer[], p_request_data varchar, p_max_row_count integer default null)
returns TABLE(cache_id bigint, data_row varchar)
 LANGUAGE plpgsql
as $function$
/*
 * select * from l3d.export_cache(array[67462],'cache:{}, nodes:{firstNodeId:0}, aux_nodes:{firstNodeId:0}')
 * select * from l3d.export_cache(array[67669],'"nodes":{}, "aux_nodes":{},"max_row_count" :50')
 *  select * from l3d.export_cache(array[67657],'')
 * 	-- select * from l3d.export_cache('/201af4c330f3ef65-67cd02ca-43a1-ade3/0/5910/data/export','{"cache":{},"nodes" :{}, "aux_nodes":{}, "shapes" :{}, "max_row_count":1000}' )
 * 	-- select * from l3d.export_cache('/201af4c330f3ef65-67cd02ca-43a1-ade3/0/5910/data/export','{"cache":{},"nodes" :{}, "aux_nodes":{}}' ) 
 *  	-- select * from l3d.export_cache('/201af4c330f3ef65-67cd02ca-43a1-ade3/0/5910/data/export','{"aux_nodes":{}}' )
 * */
declare

	v_is_cache_info_requested boolean;
	v_are_nodes_requested boolean;
	v_are_aux_nodes_requested boolean;
	v_are_shapes_requested boolean;
	v_all_data_requested boolean := true ;

	v_tables_cache_info varchar[];

	v_first_node_id integer := 0;
	v_tables_nodes varchar[];

	v_first_aux_node_id integer := 0;
	v_tables_aux_nodes varchar[];


	v_first_shape_id integer := 0; 
	v_tables_shapes varchar[];

	v_max_row_count integer; 
		
begin
	
	
		if p_request_data is not null then
	
			select		not cache_info_requested and not nodes_requested and not aux_nodes_requested and not shapes_requested,
						cache_info_requested,
						nodes_requested, 
						aux_nodes_requested,
						shapes_requested, 
						string_to_array(a.tables_cache_info, ',') tables_cache_info,
						
						coalesce( a.last_node_id::integer + 1 , 0 ) first_node_id,
					   	string_to_array( a.tables_nodes, ',' ) tables_nodes,
					   	
					   	coalesce( a.last_aux_node_id::integer + 1, 0 ) first_aux_node_id,					   	
					   	string_to_array( a.tables_aux_nodes, ',' ) tables_aux_nodes,

					   	coalesce( a.last_shape_id::integer + 1, 0 ) first_shape_id,					   						   						   
					   	string_to_array( a.tables_shapes, ',' ) tables_shapes, 
					   	
					   	max_row_count
					   	
			into 	v_all_data_requested,
					v_is_cache_info_requested, v_are_nodes_requested, v_are_aux_nodes_requested, v_are_shapes_requested, v_tables_cache_info, 
					v_first_node_id, v_tables_nodes, 
					v_first_aux_node_id, v_tables_aux_nodes, 
					v_first_shape_id, v_tables_shapes,
					v_max_row_count
			from
			(
				select	cache_info is not null		cache_info_requested,
						nodes_info is not null 		nodes_requested,
						aux_nodes_info is not null	aux_nodes_requested,
						shapes_info is not null		shapes_requested,
						substring( cache_info, '"?[tT]ables"?[\s]*:[\s]*[[]([a-zA-Z",_]+)[]]' ) 		tables_cache_info,
						
						substring( nodes_info, '"?[lL]ast[eE]xt[Ii]d"?[\s]*:[\s]*([0-9]+)' )			last_node_id,												
						substring( nodes_info, '"?[tT]ables"?[\s]*:[\s]*[[]([a-zA-Z",_]+)[]]' ) 		tables_nodes,
						
						substring( aux_nodes_info, '"?[lL]ast[eE]xt[Ii]d"?[\s]*:[\s]*([0-9]+)' )		last_aux_node_id,
						substring( aux_nodes_info, '"?[tT]ables"?[\s]*:[\s]*[[]([a-zA-Z",_]+)[]]' ) 	tables_aux_nodes,
						
						substring( shapes_info, '"?[lL]ast[eE]xt[Ii]d"?[\s]*:[\s]*([0-9]+)' )			last_shape_id,						
						substring( shapes_info, '"?[tT]ables"?[\s]*:[\s]*[[]([a-zA-Z",_]+)[]]' ) 		tables_shapes,
						
						max_row_count::integer max_row_count
						
				from
				(			
					select	substring(p_request_data, '"cache"[\s]*:[\s]*[{]([^}]*)') as cache_info,
							substring(p_request_data, '"nodes?"[\s]*:[\s]*[{]([^}]*)') as nodes_info,
							substring(p_request_data, '"aux_nodes?"[\s]*:[\s]*[{]([^}]*)') as aux_nodes_info,
							substring(p_request_data, '"shapes?"[\s]*:[\s]*[{]([^}]*)') as shapes_info,
							substring(p_request_data, '"max_row_count"[\s]*:[\s]*([0-9]+)') as max_row_count
				) a
			) a;
		
 
		
		end if;
	
		return query 	
		select a.cache_id, ( '"' || a.table_name || '":[' || array_to_string( a.data_row , ',' ) || ']' || case when v_max_row_count is not null then (',"max_row_count":' || v_max_row_count) else '' end )::varchar
		from l3d.export_cache_data( p_caches_ids, v_tables_cache_info ) a
		where v_is_cache_info_requested or v_all_data_requested
		union all
		select a.cache_id, ('"' || a.table_name || '":["' || array_to_string( a.table_data , '","' )  || '"]' || 
								case when a.table_name = 'node' then ( ',"pending_nodes":"' || a.pending_data || '"' ) else '' end
							)::varchar 	
		from l3d.export_nodes_data( p_caches_ids, v_first_node_id, v_max_row_count) a 
		where v_are_nodes_requested or v_all_data_requested
		union all
		select a.cache_id, ( '"' || a.table_name || '":["' || array_to_string( a.table_data , '","' ) || '"]' || 
								case when a.table_name = 'aux_node' then ( ',"pending_aux_nodes":"' || a.pending_data || '"' ) else '' end
						)::varchar 	
		from l3d.export_aux_nodes_data( p_caches_ids, v_first_aux_node_id, v_max_row_count) a
		where v_are_aux_nodes_requested or v_all_data_requested
		union all
		select a.cache_id,  ( '"' || a.table_name || '":["' || array_to_string( a.table_data , '","' ) || '"]'  ||
										case when a.table_name = 'shape' then ( ',"pending_shapes":"' || a.pending_data || '"' ) else '' end
						)::varchar 	
		from l3d.export_shapes_data( p_caches_ids, v_first_shape_id, v_max_row_count) a
		where v_are_shapes_requested or v_all_data_requested
		;
		
end ;
$function$;


alter function l3d.export_cache(integer[], varchar, integer)  OWNER TO i3dhub;

drop function if exists l3d.export_cache_data(integer[],varchar[], integer);

CREATE OR REPLACE FUNCTION l3d.export_cache_data(p_caches_ids integer[], p_tables varchar[] DEFAULT NULL)
 RETURNS TABLE(cache_id bigint, table_name varchar, data_row varchar[])
 LANGUAGE plpgsql
AS $function$
	/**
	 *  select * from l3d.export_cache_data(array[67462])
	 * 	 select * from l3d.export_cache_data(array[67687])
	 */ 
begin
	
	return query 
	select ic.cache_id , 'index_cache'::varchar as table_name, 
				array[						
					 '"' || coalesce( ic.url , ' ' ) ||  '"'
					, '"' || coalesce( ic.state , ' ' ) ||  '"'
					, '"' || coalesce( ic.transcoder_group , ' ' )  ||  '"'	
					, coalesce( ic.bytes , '0' )					
					, '"' || coalesce( ic.l3dgen_version, ' ' )  ||  '"'						
					, coalesce( ic.data_layout_version, '0' )	
					, coalesce( replace( ic.original_request::varchar , '\/', '/' ), '{}' )
				]::varchar[]
	from l3d.index_cache ic 	
		where ( p_tables is null or 'index_cache' ilike any(p_tables) ) and ic.cache_id = any( p_caches_ids )	
	union all
	select ids.cache_id , 'index_source' as table_name, 
				array[
						'"' || coalesce( ids.uri, ' ' ) || '"'
					,	'"' || coalesce( ids.url, ' ' ) || '"'
					,	'"' || coalesce(ids.auth_url, ' ' ) || '"'
					,	'"' || coalesce( array_to_string(ids.mime_type, ';') , ' ' ) || '"'					
					,	'"' || coalesce(ids.property , ' ' ) || '"'					
					,	'"' || coalesce(trim( ids.e_tag , '"' ) , ' ' ) || '"'					
				]::varchar[]
	from l3d.index_source ids
		where ( p_tables is null or 'index_source' ilike any(p_tables) ) and ids.cache_id = any( p_caches_ids )
	union all 
	select idst.cache_id ,'index_stats' as table_name, 
			array[
					'"' || coalesce(idst.queued_date::varchar , ' ' )  || '"'			
				,	coalesce(idst.num_reads , '0' )				
				,	'"' || coalesce(idst.last_modified_date::varchar , ' ' )	 || '"'			
				,	'"' || coalesce(idst.last_read_date::varchar , ' ' )	 || '"'			
				,	'"' || coalesce(idst.processing_start_date::varchar , ' ' )	 || '"'			
				,	'"' || coalesce(idst.processing_end_date::varchar , ' ' )  || '"'				
			]::varchar[] 
	from l3d.index_stats idst
		where ( p_tables is null or 'index_stats' ilike any(p_tables) ) and idst.cache_id = any( p_caches_ids )		
	union all		
	select ip.cache_id , 'index_properties' as table_name, array[ properties ::varchar  ]
	from l3d.index_properties ip
		where ( p_tables is null or 'index_properties' ilike any(p_tables) ) and ip.cache_id = any( p_caches_ids )
 				
	union all 	
	select ds.cache_id , 'data_stats' as table_name, array[ left(right( ( stats ::varchar ), -1) , -1) ]
	from l3d.data_stats ds 
		where ( p_tables is null or 'data_stats' ilike any(p_tables) ) and ds.cache_id = any( p_caches_ids )		
	;
		
end ;
$function$;

alter function l3d.export_cache_data(integer[],varchar[])  OWNER TO i3dhub;

drop function if exists l3d.export_shapes_data(integer[], integer, integer);
CREATE OR REPLACE FUNCTION l3d.export_shapes_data(p_caches_ids integer[], p_offset integer DEFAULT 0, p_limit integer DEFAULT NULL::integer)
 RETURNS TABLE(cache_id integer, table_name character varying, table_data character varying[], pending_data boolean)
 LANGUAGE plpgsql
AS $function$
declare
	/*
 *   select * from l3d.export_shapes_data(array[67462], array['shape','shape_instance','face_edge_instance'])
 * 	select * from l3d.export_shapes_data(array[67687], 0, null)
 * 
 * */
		v_return integer;
--	 v_t0 timestamp := clock_timestamp();
begin
		
	return query
	select 	a.cache_id , 
			a.table_name,
			a.table_data, 
			case when a.table_name = 'shape' then sum(a.pending_data) over() <> 0 end as pending_data		
	from
	(
		select a.cache_id ,  
			  'shape'::varchar as table_name, 
				array_agg( '[' || array_to_string( a.shape_data , ']-,-[') || ']' ) ::varchar[] as table_data ,
				sum(a.pending_data)  as pending_data
		from
		(
			select a.cache_id ,  
					array[ a.ext_shape_id::varchar, array_to_string( coalesce( a.edges , array[]::integer[] ), ',' ), array_to_string( coalesce( a.faces , array[]::integer[] ) , ',' ) ] as shape_data ,
					pending_edges + pending_faces as pending_data
			from
			(
				select sha.cache_id, sha.ext_shape_id, 
					( select array_agg( e.ext_edge_id ) from (select e.ext_edge_id from l3d.edge e where e.cache_id = sha.cache_id and e.shape_id = sha.id order by e.ext_edge_id limit rng.l offset rng.o ) e ) edges,
					( select array_agg( f.ext_face_id ) from (select f.ext_face_id from l3d.face f where f.cache_id = sha.cache_id  and f.shape_id = sha.id order by f.ext_face_id limit rng.l offset rng.o) f ) faces,	
					( select case when rng.l is not null and array_agg( e.ext_edge_id ) is not null then 1 else 0 end from (select e.ext_edge_id from l3d.edge e where e.cache_id = sha.cache_id and e.shape_id = sha.id order by e.ext_edge_id limit 1 offset ( rng.o + rng.l ) ) e ) pending_edges,
					( select case when rng.l is not null and array_agg( f.ext_face_id ) is not null then 1 else 0 end from (select f.ext_face_id from l3d.face f where f.cache_id = sha.cache_id  and f.shape_id = sha.id order by f.ext_face_id limit 1 offset ( rng.o + rng.l ) ) f ) pending_faces				   					   
				from l3d.shape sha , ( select p_limit as l, p_offset as o) rng		
				where sha.cache_id  = any( p_caches_ids)	 
		   ) a
		) a
			where shape_data is not null
		
		group by a.cache_id		

		union all

		--		[ext_shap_instance_id]-,-[ext_shape_id]-,-[ext_node_id]
	
		select a.cache_id , 'shape_instance'::varchar as table_name, array_agg( a.shape_instance_data) as table_data, 
				sum( case when a.idx = 1 and a.max_idx > p_limit then 1 else 0 end )  as pending_data
		from
		(
		
			select  a.cache_id , ( '[' || array_to_string( a.shape_instance_data , ']-,-[' )  || ']')::varchar as shape_instance_data, a.idx,  max(a.idx) over() as max_idx
			from
			(
				select shi.cache_id, array[ shi.shape_instance_id , sha.ext_shape_id , nod.ext_node_id ]::varchar[]  shape_instance_data , row_number() over() idx
				from l3d.shape_instance shi, l3d.shape sha, l3d.node nod
				where 
					shi.cache_id = any( p_caches_ids )
				and sha.id = shi.shape_id 
				and shi.node_id = nod.id							 
				limit ( p_limit + 1) offset p_offset
			) a
				where p_limit is null or a.idx <= p_limit
		) a
		group by a.cache_id
	
		union all
	
		select a.cache_id, 'face_instance'::varchar as table_name , array_agg(  a.face_instance_data ) as table_data , 
					sum( case when a.idx = 1 and a.max_idx > p_limit then 1 else 0 end )  as pending_data 
		from
		(
			select a.cache_id , ( '[' || array_to_string( a.face_instance_data , ']-,-[') || ']' )::varchar face_instance_data, a.idx , max(a.idx) over() as max_idx
			from
			(		
				select	fai.cache_id, 
						array['f', fac.ext_face_id::varchar, shi.shape_instance_id::varchar, coalesce( fito.original_face_id::varchar , '') ]  as face_instance_data , 
						row_number() over() idx
				from l3d.face_instance fai  
					join  l3d.face fac on fac.id = fai.face_id
					join l3d.shape_instance shi on shi.id = fai.shape_instance_id
					left join l3d.face_instance_to_original fito on fito.face_instance_id = fai.id
				where 
					fai.cache_id = any(p_caches_ids) 
				offset p_offset limit ( p_limit + 1 )
			) a
		)  a
			where p_limit is null or a.idx <= p_limit
		group by a.cache_id
		
		union all
	
		select a.cache_id, 'edge_instance'::varchar as table_name , 
					array_agg( ( '[' || array_to_string( a.edge_instance_data , ']-,-[') || ']' )::varchar ) as table_data , 
					sum( case when a.idx = 1 and a.max_idx > p_limit then 1 else 0 end ) as pending_data 
		from
		(
			select a.cache_id , a.edge_instance_data, a.idx , max(a.idx) over() as max_idx
			from
			(		
				select edi.cache_id, array[ 'e' , edg.ext_edge_id::varchar,  shi.shape_instance_id::varchar ] as edge_instance_data, row_number() over() idx
				from l3d.edge_instance edi , l3d.edge edg, l3d.shape_instance shi
				where	
					edi.cache_id = any(p_caches_ids) 
				and edg.id = edi.edge_id
				and shi.id = edi.shape_instance_id
				offset p_offset limit ( p_limit + 1 )
			) a
		) a
			where  p_limit is null or a.idx <= p_limit
		group by a.cache_id

	)  a 
	;

--	raise warning 'export_shapes_data, time %' , (clock_timestamp() - v_t0 );

end;
$function$;
alter function l3d.export_shapes_data(integer[], integer, integer)  OWNER TO i3dhub;


drop function if exists l3d.export_nodes_data(integer[],integer,integer,varchar[]);

CREATE OR REPLACE FUNCTION l3d.export_nodes_data(p_caches_ids integer[], p_first_ext_node_id integer DEFAULT NULL::integer, p_max_nodes_count integer DEFAULT NULL::integer, p_tables character varying[] DEFAULT NULL::character varying[])
 RETURNS TABLE(cache_id integer, table_name character varying, table_data character varying[], pending_data boolean)
 LANGUAGE plpgsql
AS $function$
declare
	/** 
	 *  select * from l3d.export_nodes_data(76008, null, 1000, array['node','metadata'])
	 * 
	 *  select * from l3d.export_nodes_data(275, null, 1000)
	 *  select * from l3d.export_nodes_data(78263, null, 1000)
	 * 
	 * select * from l3d.export_nodes_data(array[67687], 0, 1000)
	 * 
	 */
	v_min_node_id integer := coalesce( p_first_ext_node_id, 0 );
	v_max_node_id integer := v_min_node_id +  p_max_nodes_count - 1; 
begin
							
	return query
	select a.cache_id, a.table_name, a.table_data , 
			case when a.table_name = 'node' then
			 	  p_max_nodes_count is not null and p_max_nodes_count = array_length( a.table_data, 1 ) and exists ( select 1 from l3d.node n where n.cache_id = a.cache_id and n.ext_node_id > v_max_node_id )		
			end as pending_data	
	from
	(
			select a.cache_id, 'node'::varchar as table_name, array_agg( ( '[' || array_to_string( array[a.node_id, a."label", a.parent_node_id], ']-,-[') || ']' )::varchar ) as table_data
			from 
			(
				select		nod.cache_id , 
							nod.ext_node_id::varchar as node_id, 
					   		case when nod."label" ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(nod.label, -1), -1) else nod."label" end as "label",
					   		nod.parent_node_id::varchar as parent_node_id
				from l3d.node nod
				where 
					( p_tables is null or 'node' ilike any( p_tables ) ) 			and 
					nod.cache_id  = any( p_caches_ids )								and 
					( p_max_nodes_count is null or nod.ext_node_id between v_min_node_id and v_max_node_id )						
			) a	
			group by a.cache_id
			
			union all 
			
			select a.cache_id, 'metadata'::varchar as table_name, array_agg( ( '[' || array_to_string( array[a.value_type, a.key, coalesce(a.value_num, a.value_str), a.node_id], ']-,-[') || ']' )::varchar ) as table_data	
			from
			(
				select	a.cache_id, 
					a.node_id, 
						a."key"  as "key",
						case when a.value_num is not null then 'n' when a.value_str is not null then 's'  end as value_type,
						a.value_num as value_num,
						a.value_str as value_str
		/*				
					   case 				   		
					   		when a.value_str !~ '[^[:alnum:].]+' then a.value_str 
					   		else left( right( array_to_json( array[a.value_str] )::varchar, -2 ), -2 ) 
					   	end	as value_str
		*/			   	
				from
				(
					select	nod.cache_id , nod.ext_node_id::varchar as node_id,
							case when mda."key" ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(mda."key", -1), -1) else mda."key" end as "key",										
							mda.value_num ::varchar as value_num,
							case when mda.value_str ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(mda.value_str, -1), -1) else mda.value_str end as value_str 											
					from l3d.node nod, l3d.metadata mda 
					where 
						( p_tables is null or 'metadata' ilike any(p_tables) ) 	and
						nod.cache_id = any( p_caches_ids )						and 
						( p_max_nodes_count is null or nod.ext_node_id between v_min_node_id and v_max_node_id )  and
						mda.node_id = nod.id
				) a	
			) a
			group by a.cache_id
			
			union all 
			
			select a.cache_id , 'tag'::varchar as table_name, array_agg( ( '[' || array_to_string( array[a."name", a.value, a.node_id], ']-,-[' ) || ']' )::varchar ) as table_data
			from 
			(
				select	a.cache_id, a.node_id, 
						a."name" as "name",
						a.value as value
		/*				
					   case 				   
					   		when a.value !~ '[^[:alnum:].]+' then a.value
					   		else left( right( array_to_json( array[a.value] )::varchar,-2 ),-2 ) 
					   	end	as value
		*/			   	
				from
				(
					select	nod.cache_id , nod.ext_node_id::varchar as node_id,
							case when tn."name" ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(tn."name", -1), -1) else tn."name" end as "name",										
							case when tv.value ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(tv.value, -1), -1) else tv.value end as value 											
					from l3d.node nod, l3d.tags_values tv , l3d.tag_name tn 
					where 
						( p_tables is null or 'tag' ilike any(p_tables) ) 	and	
						nod.cache_id = any( p_caches_ids )	 				and 
						( p_max_nodes_count is null or nod.ext_node_id between v_min_node_id and v_max_node_id ) and
						tv.node_id = nod.id 								and 
						tn.id  = tv.tag_name_id								
				) a
			) a		
			group by a.cache_id			
	) a
	where a.table_data is not null;
	
		
end;
$function$;

	
ALTER FUNCTION l3d.export_nodes_data(integer[],integer,integer,varchar[]) OWNER TO i3dhub;

drop function if exists l3d.export_aux_nodes_data(integer[],integer,integer,varchar[]);
drop function if exists l3d.export_aux_nodes_data(integer[],integer,integer,character varying[]);


CREATE OR REPLACE FUNCTION l3d.export_aux_nodes_data(p_caches_ids integer[], p_first_ext_node_id integer DEFAULT NULL::integer, p_max_nodes_count integer DEFAULT NULL::integer, p_tables character varying[] DEFAULT NULL::character varying[])
 RETURNS TABLE(cache_id integer, table_name character varying, table_data character varying[], pending_data boolean)
 LANGUAGE plpgsql
AS $function$
declare
	/**
	 *  
	 *   select * from l3d.data_stats where cache_id = 67657
	 * 	select * from l3d.export_aux_nodes_data(array[67683], 0, 1000)
	 *   
	 * 	select * from l3d.export_aux_nodes_data(277)
	 * 
	 * 	 * 	select * from l3d.export_aux_nodes_data(435)
	 */
	v_min_node_id integer := coalesce( p_first_ext_node_id, 0);
	v_max_node_id integer := v_min_node_id + p_max_nodes_count - 1; 
--	 v_t0 timestamp := clock_timestamp();
begin
	

	
	return query
	select a.cache_id, a.table_name, a.table_data,
			case when a.table_name = 'aux_node' then
			 	  p_max_nodes_count is not null and p_max_nodes_count = array_length( a.table_data, 1 ) and exists ( select 1 from l3d.aux_node n where n.cache_id = a.cache_id and n.cache_id = any( p_caches_ids ) and n.ext_aux_node_id > v_max_node_id )			
			end as pending_data	
	from 
	(

		select a.cache_id, 'aux_node'::varchar as table_name, 
				array_agg( ( '[' || array_to_string( array[ a.node_id, a.node_type, a."label", a.face_links ] , ']-,-[') || ']' )::varchar ) as table_data 
		from
		(
			select a.cache_id,
					a.node_id, 
					case when a.label !~ '[^[:alnum:].]+' then a.label else left( right( array_to_json( array[a.label] )::varchar, -2 ), -2 ) end as "label",
					a.node_type as node_type,
					(
						select  coalesce( array_to_string( array_agg( face_link ), ',' ) , '')   from l3d.foreign_face_link where aux_node_id =  a.id
					) as face_links
					
			from
			(
				select  nod.id, nod.cache_id , nod.ext_aux_node_id::varchar as node_id,		
						case when nod."label" ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(nod.label, -1), -1) else nod."label" end as "label",
						case when ant."name" ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(ant."name", -1), -1) else ant."name" end as node_type
				from l3d.aux_node nod, l3d.aux_node_type ant 
					where 
						( p_tables is null or 'aux_node' ilike any( p_tables ) ) 		and 
						nod.cache_id = any( p_caches_ids )								and 
						(p_max_nodes_count is null or nod.ext_aux_node_id between v_min_node_id and v_max_node_id) 	and
						ant.id = nod.aux_node_type_id				
			) a	
		) a
		group by a.cache_id

		union all 

		select a.cache_id, 'aux_node_to_aux_node'::varchar as table_name, array_agg( ( '[' || array_to_string( array[ a.pred_id, a.succ_ids ] , ']-,-[') || ']' )::varchar ) as table_data
		from 
		(
			select  pred.cache_id, pred.ext_aux_node_id::varchar as pred_id, array_to_string( array_agg( succ.ext_aux_node_id ) , ',') succ_ids
			from l3d.aux_node succ , l3d.aux_node_to_aux_node ntn , l3d.aux_node pred
			where 
				( p_tables is null or 'aux_node_to_aux_node' ilike any( p_tables ) )						and 
				pred.cache_id = any( p_caches_ids )															and 
				(p_max_nodes_count is null or pred.ext_aux_node_id between v_min_node_id and v_max_node_id)	and			
				ntn.succ_id = succ.id																		and
				ntn.pred_id = pred.id
			group by pred.cache_id, pred.ext_aux_node_id
		) a
		group by a.cache_id

		union all 
		
		select a.cache_id, 'aux_value'::varchar as table_name, array_agg( ( '[' || array_to_string( array[a.node_id, a.value_type, a."name", coalesce(a.value_num, a.value_str )], ']-,-[') || ']' )::varchar ) as table_data 
		from
		(
			select	a.cache_id,
					a.node_id, 
					a."name" as "name",
					case when a.value_num is not null then 'n' when a.value_str is not null then 's'  end as value_type,
					a.value_num ::varchar as value_num,
					case when a.value_str !~ '"' then a.value_str else replace( replace( a.value_str , '\"', '"'), '"' , '\"') end as value_str
			from
			(
				select	nod.cache_id , nod.ext_aux_node_id::varchar as node_id,
						avn.name,			
						av.value_num ::varchar as value_num,
						case when av.value_str ~ '^((\\")|"|'')(.*)((\\")|"|'')$' then left(right(av.value_str, -1), -1) else av.value_str end as value_str
				from l3d.aux_node nod, l3d.aux_value av , l3d.aux_value_name avn 
				where 
					( p_tables is null or 'aux_value' ilike any(p_tables) )	 	and 
					nod.cache_id = any(p_caches_ids)							and 
						(p_max_nodes_count is null or nod.ext_aux_node_id between v_min_node_id and v_max_node_id) 	and
					av.aux_node_id = nod.id 									and		
					avn.id = av.aux_value_name_id 						
			) a	
		) a	
		group by a.cache_id		

		union all 

		select a.cache_id, 'aux_attributes'::varchar as table_name, array_agg( ( '[' || array_to_string( array[a.node_id, a."key", a.value], ']-,-[') || ']' )::varchar ) as table_data
		from 
		(
			select	a.cache_id, 
					a.node_id, 
					a."key" as "key",
					case when a.value !~ '"' then a.value else replace( replace( a.value , '\"', '"'), '"' , '\"') end as value 
			from
			(
				select	nod.cache_id , nod.ext_aux_node_id::varchar as node_id,
						aa.key,										 											
						aa.value
				from l3d.aux_node nod, l3d.aux_attributes aa  
				where 
					( p_tables is null or 'aux_attributes' ilike any(p_tables) )	and 
					nod.cache_id = any(p_caches_ids) 								and 
					(p_max_nodes_count is null or nod.ext_aux_node_id between v_min_node_id and v_max_node_id) 	and
					aa.aux_node_id = nod.id 			
			) a
		) a
		group by a.cache_id	

				
		union all 

		select a.cache_id, 'edge_instance_to_aux_node'::varchar as table_name, 
				array_agg( ( '[' || array_to_string( array[ 'e', a.edge_id, a.shape_instance_id, a.aux_node_id], ']-,-[') || ']' )::varchar ) as table_data	
		from 
		(	
			select 	aux.cache_id, 
					aux.ext_aux_node_id::varchar as aux_node_id ,
					ed.ext_edge_id::varchar as edge_id, 
					sh.ext_shape_id::varchar  as shape_id,
					si.shape_instance_id::varchar as shape_instance_id,
					nod.ext_node_id::varchar  as node_id
			from 	l3d.aux_node aux, 
					l3d.edge_instance_to_aux_node eitan , 
					l3d.edge_instance ei ,
					l3d.edge ed ,
					l3d.shape sh, 
					l3d.shape_instance si ,
					l3d.node nod
			where 
				( p_tables is null or 'edge_instance_to_aux_node' ilike any(p_tables) )						and 
				aux.cache_id = any(p_caches_ids) 															and 	
				(p_max_nodes_count is null or aux.ext_aux_node_id between v_min_node_id and v_max_node_id) 	and										
				eitan.aux_node_id = aux.id 																	and	
				ei.id = eitan.edge_instance_id 																and
				ed.id = ei.edge_id 																			and
				sh.id = ed.shape_id 																		and
				si.id = ei.shape_instance_id 																and 
				nod.id = si.node_id
		) a
		group by a.cache_id		
	
		union all
	
		select a.cache_id, 'face_instance_to_aux_node'::varchar as table_name, 
				array_agg( ( '[' || array_to_string( array['f', a.face_id, a.shape_instance_id, a.aux_node_id], ']-,-[') || ']' )::varchar ) as table_data
		from 
		(
			select  aux.cache_id, 
					aux.ext_aux_node_id::varchar as aux_node_id ,
					fc.ext_face_id::varchar as face_id, 
					sh.ext_shape_id::varchar  as shape_id,
					si.shape_instance_id::varchar as shape_instance_id,
					nod.ext_node_id::varchar  as node_id
			from 	l3d.aux_node aux, 
					l3d.face_instance_to_aux_node fitan , 
					l3d.face_instance fi ,
					l3d.face fc ,
					l3d.shape sh, 
					l3d.shape_instance si ,
					l3d.node nod
			where 
				( p_tables is null or 'face_instance_to_aux_node' ilike any(p_tables) )						and 
				aux.cache_id = any(p_caches_ids) 															and 	
				(p_max_nodes_count is null or aux.ext_aux_node_id between v_min_node_id and v_max_node_id) 	and								
				fitan.aux_node_id = aux.id 																	and	
				fi.id = fitan.face_instance_id  															and
				fc.id = fi.face_id 																			and				
				sh.id = fc.shape_id 																		and
				si.id = fi.shape_instance_id																and
				nod.id = si.node_id	 
		) a	
		group by a.cache_id


	) a 
	where a.table_data is not null
	;	

--	raise warning 'export_aux_nodes_data, time %' , (clock_timestamp() - v_t0 );
end;
$function$;

ALTER FUNCTION l3d.export_aux_nodes_data(integer[],integer,integer,character varying[]) OWNER TO i3dhub;


drop FUNCTION IF exists l3d.import_cache(varchar, varchar, boolean);
CREATE OR REPLACE FUNCTION l3d.import_cache(p_uri character varying, p_data character varying, p_flush_data boolean DEFAULT false)
 RETURNS TABLE(o_result character varying)
 LANGUAGE plpgsql
AS $function$
declare
/*
 * 
 *   perform l3d.import_cache('"cache":{"5b2f17fa79e7cdc3-45d38cf7-4645-b994","0","275"},"node":{"[0]-,-[sourcefile]","[127]-,-[Curveset000]","[128]-,-[PRE-MACHINED_CRANKCASE]","[1]-,-[Section Geometry]","[2]-,-[5_78_3_Motortraeger_N1-_2_1]"},"metadata":{"[n]-,-[IGDNodeID]-,-[0]-,-[0]","[n]-,-[JTPath]-,-[16777200]-,-[0]","[n]-,-[REVISION]-,-[29]-,-[0]","[n]-,-[j0CPArea]-,-[2.34474]-,-[0]","[n]-,-[j0CentreOfGravity_x]-,-[10.6406]-,-[0]","[n]-,-[j0CentreOfGravity_y]-,-[-17.6076]-,-[0]","[n]-,-[j0CentreOfGravity_z]-,-[91.1588]-,-[0]","[n]-,-[j0Density]-,-[2750]-,-[0]","[n]-,-[j0DqNXPrtVrs]-,-[1509]-,-[0]","[n]-,-[j0GeoVolume]-,-[0.0117051]-,-[0]","[n]-,-[j0Height]-,-[0]-,-[0]","[n]-,-[j0Ixx]-,-[0.411792]-,-[0]","[n]-,-[j0Ixy]-,-[0.0441772]-,-[0]","[n]-,-[j0Ixz]-,-[0.0199545]-,-[0]","[n]-,-[j0Iyy]-,-[1.18057]-,-[0]","[n]-,-[j0Iyz]-,-[-0.0222218]-,-[0]","[n]-,-[j0Izz]-,-[1.18413]-,-[0]","[n]-,-[j0WeightCalc]-,-[32.189]-,-[0]","[s]-,-[3D-Master]-,-[Powertrain]-,-[0]","[s]-,-[ACTIVELAYERFILTER]-,-[FINAL_PART]-,-[0]","[s]-,-[All_UFI]-,-[18001, 18002, 18003, 18004, 18005, 18006, 16801, 16802, 16803, 16804, 16805, 16806, 16807, 16808, 16809, 16810, 16811, 16812, 16813, 16814, 18201, 18202, 18203, 18204, 18205, 18206, 18207, 18208, 18209, 18210, 18211, 18212, 18213, 18214, 18215, 16815, 14701, 14702, 14703, 11301, 11302, 11303, 11304, 11305, 11306, 11501, 11502, 11201, 11202, 11203, 11204, 11307, 12001, 12002, 11101, 11001, 22001, 22002, 22003, 22004, 22005, 22006, 22007, 22008, 22009, 22010, 22011, 22012, 22013, 22014, 22015, 22016, 22017, 22018, 21801, 21802, 21803, 21804, 21301, 21302, 21303, 28401, 28402, 28403, 28404, 28405, 28406, 28407, 28408, 28409, 28410, 28411, 28412, 28413, 28414, 28601, 25001, 25101, 25102, 25103, 25104, 25105, 25106, 25107, 25108, 25109, 25110, 25111, 24701, 24702, 24703, 24704, 24705, 24706, 24707, 28101, 28102, 28103, 28104, 28105, 28106, 28107, 28108, 28109, 28110, 28111, 28112, 21101, 21102, 21103, 21104, 21105, 21106, 25201, 25202, 25203, 25204, 25205, 25206, 25207, 25208, 25209, 25210, 25211, 25212, 27501, 27502, 27503, 27504, 27505, 27506, 27507, 27508, 27509, 27510, 27511, 27512, 27513, 27514, 25801, 25802, 25803, 25804, 25805, 25806, 25807, 22801, 22802, 22803, 22804, 22805, 22806, 22807, 22808, 22809, 22810, 22811, 27401, 27402, 27403, 27404, 27405, 27406, 27407, 27408, 27409, 27410, 27411, 27412, 27413, 27414, 27601, 27602, 27603, 27604, 27605, 27606, 27607, 27608, 27609, 27610, 27611, 27612, 27613, 27614, 21001, 31401, 31402, 31403, 31404, 31601, 31602, 31603, 31604, 31701, 32001, 32002, 32003, 32004, 32005, 32006, 32007, 34001, 34002, 34801, 34802, 32801, 31101, 31102, 31103, 31104, 31105, 31001, 31106, 31107, 33901, 36101, 36102, 36103, 36104, 36105, 36106, 36107, 36108, 36109, 33701, 33702, 33703, 33704, 33705, 41201, 41202, 41501, 41502, 41503, 43901, 43902, 43903, 42001, 42002, 42003, 42004, 42005, 42006, 42007, 42008, 42009, 42010, 42011, 42012, 42013, 42014, 42015, 42016, 44101, 44102, 44103, 45001, 45002, 45201, 45202, 45601, 45602, 45603, 45604, 45003, 44701, 44702, 45203, 44703, 45101, 41001, 41101, 41102, 41103, 41104, 41105, 42101, 42102, 42103, 42104, 42105, 42106, 42107, 42108, 51501, 51502, 56501, 52001, 52002, 52003, 52004, 52005, 52006, 52007, 52008, 54501, 54502, 52009, 53401, 53402, 53403, 53301, 53302, 53303, 53304, 53305, 53306, 54901, 54902, 54903, 54904, 54905, 54906, 54907, 54908, 54909, 52010, 56301, 55601, 55602, 55603, 54701, 54702, 54703, 54704, 56101, 57801, 57802, 57803, 57901, 54801, 54802, 54803, 54804, 54805, 54806, 54807, 55701, 51001, 51101, 51102, 51103, 51104, 51105, 51106, 51107, 51108, 51109, 51110, 51111, 51112, 51113, 51114, 51115, 51116, 51117, 51118, 51119, 51120, 51121, 51122, 51123, 51124, 51125, 51126, 51127, 51128, 51129, 51130, 51131, 51132, 51133, 51134, 51135, 51136, 65901, 65902, 65903, 65904, 65905, 65906, 65907, 65908, 67801, 67802, 67901, 67902, 64101, 64102, 63101, 63401, 63402, 63501, 63102, 65101, 65102, 66601, 66602, 62001, 62002, 62003, 62004, 62005, 62006, 62007, 62008, 61101, 61102, 61103, 61104, 61105, 61106, 61107, 61108, 61001, 61109, 61110, 61111, 61112, 61113, 61114, 61115, 61116, 61117, 61118, 65909, 66603, 24801, 22101, 22102, 22103, 22104, 27301, 27302, 27701, 27702, 27703, 27704, 27705, 27706, 27707, 27708, 27709, 27710, 27711, 27712, 27713, 27714, 27715, 27716, 27717, 27718, 27719, 27720, 27721, 27722, 27723, 47101, 47102, 47103, 47104, 47105, 47106, 47107, 16501, 16502, 16503, 16504, 16505, 16506, 16507, 16508, 11102, 11103, 15601, 15602, 15603, 15604, 13901, 25701, 54705, 61119, 42109, 36110, 61120, 22019, 22020, 22021]-,-[0]","[s]-,-[Business-Unit]-,-[MBC]-,-[0]","[s]-,-[CAD Source]-,-[NX]-,-[0]","[s]-,-[Configuration_Name]-,-[Single Model (SM, SML)]-,-[0]","[s]-,-[Contents]-,-[Final_Part]-,-[0]","[s]-,-[Department]-,-[Powertrain]-,-[0]","[s]-,-[FTool_Modif_Date]-,-[2015-04-21T06:38:29Z]-,-[0]","[s]-,-[FeatureGroup_extension]-,-[PTCOL]-,-[0]","[s]-,-[HQM_CHECKPROFILE_HASH]-,-[4d71965d38bce44bce40ad1f6e69fba7]-,-[0]","[s]-,-[HQM_MAPPING_HASH]-,-[78746bc93d06d6c21b4404cf7a3b092d]-,-[0]","[s]-,-[HQM_RESULT_DEF_HASH]-,-[fd05e7ad1ea5e61087c1779e2617018a]-,-[0]","[s]-,-[HQM_TOTAL_HASH]-,-[0b237b167c3d5d7bf7f46c2f7509bc94]-,-[0]","[s]-,-[JT_PROP_MEASUREMENT_UNITS]-,-[millimeters]-,-[0]","[s]-,-[LAYERFILTER000]-,-[FINAL_PART=28241305,28486178,29210577]-,-[0]","[s]-,-[LAYERFILTER001]-,-[DAI_MASS_PROPERTIES=28241305,28486178,29210577]-,-[0]","[s]-,-[Language Selection]-,-[German]-,-[0]","[s]-,-[MAPPING ID]-,-[MBC Powertrain]-,-[0]","[s]-,-[Manufacturing_Method]-,-[Casting]-,-[0]","[s]-,-[Migration]-,-[No]-,-[0]","[s]-,-[Mirror_Part]-,-[No]-,-[0]","[s]-,-[Mode Selection]-,-[Creation]-,-[0]","[s]-,-[NOMENCLATURE]-,-[ZB KURBELGEHAEUSE_PMI]-,-[0]","[s]-,-[NX_MaterialMissingAssignments]-,-[TRUE]-,-[0]","[s]-,-[NX_MaterialMultipleAssigned]-,-[TRUE]-,-[0]","[s]-,-[Name]-,-[sourcefile]-,-[0]","[s]-,-[PARTNUMBER]-,-[A2560105500_3]-,-[0]","[s]-,-[Part Group]-,-[Zyl.-Kurbelgehaeuse allg.]-,-[0]","[s]-,-[RNI_PARTTYPE]-,-[JT]-,-[0]","[s]-,-[RNI_PARTTYPE_CLASSIFICATION_PROGRAM]-,-[prt2jt 2016-04-29 NX1003MP4]-,-[0]","[s]-,-[Release_Relevant]-,-[Yes]-,-[0]","[s]-,-[Startpart_Language]-,-[English]-,-[0]","[s]-,-[Startpart_Name]-,-[daimler_startpart_model.prt]-,-[0]","[s]-,-[Startpart_Version]-,-[1.0.0.4]-,-[0]","[s]-,-[SubManufacturing_Method]-,-[Undefined]-,-[0]","[s]-,-[Supplier_Part]-,-[No]-,-[0]","[s]-,-[Translator Version]-,-[NX 10.0.3]-,-[0]","[s]-,-[Type Group]-,-[Kurbelgehaeuse]-,-[0]","[s]-,-[j0CPMatAL]-,-[]-,-[0]","[s]-,-[j0CPMatWEZ]-,-[]-,-[0]","[s]-,-[j0CPModTyp]-,-[3D-Master_PT]-,-[0]","[s]-,-[j0DqChkScheme]-,-[MBC_Powertrain]-,-[0]","[s]-,-[j0DqTestConfig]-,-[DAIMLER_COMMON_NX100_01_2016]-,-[0]","[s]-,-[j0DqTestDate]-,-[2016/06/03-14:05:32:385]-,-[0]","[s]-,-[j0DqTestSeal]-,-[j0DqSeal2]-,-[0]","[s]-,-[j0MasterDataVersion]-,-[2.0.2]-,-[0]","[s]-,-[j0Material2]-,-[AL-LEG.\tEN 1706-\tAC-46000\t\tDF]-,-[0]","[s]-,-[j0MaterialDetailMatrix]-,-[AL-LEG.\tEN 1706-\tAC-46000\t\tDF\t\t\t0.00000000\t\tNX_GENERAL\tPRE-MACHINED_CRANKCASE]-,-[0]","[s]-,-[j0QEVNumber]-,-[QEV111ACRTON]-,-[0]","[n]-,-[IGDLeafIndex]-,-[0]-,-[127]","[n]-,-[IGDNodeID]-,-[127]-,-[127]","[n]-,-[LODLabel0]-,-[0.2]-,-[127]","[n]-,-[LODLabel1]-,-[0.8]-,-[127]","[n]-,-[LODLabel2]-,-[1.8]-,-[127]","[s]-,-[AdvCompressLODLevel]-,-[0.000000 0.000000 1.000000 ]-,-[127]","[s]-,-[JTPath]-,-[16777215-29210321]-,-[127]","[s]-,-[JT_PROP_MEASUREMENT_UNITS]-,-[millimeters]-,-[127]","[s]-,-[LAYER]-,-[1,29210577]-,-[127]","[s]-,-[Name]-,-[Curveset000]-,-[127]","[s]-,-[Translation Date]-,-[19.8.2016, 6:55:49]-,-[127]","[n]-,-[IGDLeafIndex]-,-[1]-,-[128]","[n]-,-[IGDNodeID]-,-[128]-,-[128]","[n]-,-[LODLabel0]-,-[0.2]-,-[128]","[n]-,-[LODLabel1]-,-[0.8]-,-[128]","[n]-,-[LODLabel2]-,-[1.8]-,-[128]","[n]-,-[j0CPArea]-,-[2.34474]-,-[128]","[n]-,-[j0CentreOfGravity_x]-,-[10.6406]-,-[128]","[n]-,-[j0CentreOfGravity_y]-,-[-17.6076]-,-[128]","[n]-,-[j0CentreOfGravity_z]-,-[91.1588]-,-[128]","[n]-,-[j0Density]-,-[2750]-,-[128]","[n]-,-[j0GeoVolume]-,-[0.0117051]-,-[128]","[n]-,-[j0Height]-,-[0]-,-[128]","[n]-,-[j0Ixx]-,-[0.411792]-,-[128]","[n]-,-[j0Ixy]-,-[0.0441772]-,-[128]","[n]-,-[j0Ixz]-,-[0.0199545]-,-[128]","[n]-,-[j0Iyy]-,-[1.18057]-,-[128]","[n]-,-[j0Iyz]-,-[-0.0222218]-,-[128]","[n]-,-[j0Izz]-,-[1.18413]-,-[128]","[n]-,-[j0WeightCalc]-,-[32.189]-,-[128]","[s]-,-[AdvCompressLODLevel]-,-[0.000000 0.000000 1.000000 ]-,-[128]","[s]-,-[CAD_MATERIAL]-,-[AL-LEG. EN 1706- AC-46000 DF]-,-[128]","[s]-,-[JTPath]-,-[16777215-28241049]-,-[128]","[s]-,-[JT_PROP_MEASUREMENT_UNITS]-,-[millimeters]-,-[128]","[s]-,-[LAYER]-,-[1,28241305]-,-[128]","[s]-,-[Material]-,-[AL-LEG. EN 1706- AC-46200 DF]-,-[128]","[s]-,-[NX_Material]-,-[AL-LEG. EN 1706- AC-46000 DF]-,-[128]","[s]-,-[Name]-,-[PRE-MACHINED_CRANKCASE]-,-[128]","[s]-,-[Translation Date]-,-[19.8.2016, 6:55:52]-,-[128]","[s]-,-[j0CPMatAL]-,-[]-,-[128]","[s]-,-[j0CPMatWEZ]-,-[]-,-[128]","[s]-,-[j0Material2]-,-[AL-LEG.\tEN 1706-\tAC-46000\t\tDF]-,-[128]","[s]-,-[j0QEVNumber]-,-[QEV111ACRTON]-,-[128]","[n]-,-[IGDNodeID]-,-[1]-,-[1]","[s]-,-[JTPath]-,-[16777215-16777214]-,-[1]","[s]-,-[Name]-,-[Section Geometry]-,-[1]","[n]-,-[IGDNodeID]-,-[2]-,-[2]","[s]-,-[JTPath]-,-[16777215-16777214-16777213]-,-[2]","[s]-,-[Name]-,-[5_78_3_Motortraeger_N1-_2_1]-,-[2]"},"tag":{"[label]-,-[5_78_3_Motortraeger_N1-_2_1]-,-[2]","[label]-,-[Section Geometry]-,-[1]","[label]-,-[PRE-MACHINED_CRANKCASE]-,-[128]","[label]-,-[Curveset000]-,-[127]","[label]-,-[sourcefile]-,-[0]"},"aux_node":{"[4]-,-[Dimension]-,-[I3DHInternalGrouping]","[3]-,-[DatumFeatureSymbol]-,-[I3DHInternalGrouping]","[2]-,-[Chamfer]-,-[I3DHInternalGrouping]","[1]-,-[Centerline]-,-[I3DHInternalGrouping]","[0]-,-[PMI]-,-[I3DHInternalGrouping]"},"aux_value":{"[4]-,-[s]-,-[label]-,-[Dimension]","[4]-,-[n]-,-[OriginalID]-,-[-2147480000]","[4]-,-[n]-,-[OriginalTypeID]-,-[-2147480000]","[4]-,-[s]-,-[NodePath]-,-[]","[3]-,-[s]-,-[label]-,-[DatumFeatureSymbol]","[3]-,-[n]-,-[OriginalID]-,-[-2147480000]","[3]-,-[n]-,-[OriginalTypeID]-,-[-2147480000]","[3]-,-[s]-,-[NodePath]-,-[]","[2]-,-[s]-,-[label]-,-[Chamfer]","[2]-,-[n]-,-[OriginalID]-,-[-2147480000]","[2]-,-[n]-,-[OriginalTypeID]-,-[-2147480000]","[2]-,-[s]-,-[NodePath]-,-[]","[1]-,-[s]-,-[label]-,-[Centerline]","[1]-,-[n]-,-[OriginalID]-,-[-2147480000]","[1]-,-[n]-,-[OriginalTypeID]-,-[-2147480000]","[1]-,-[s]-,-[NodePath]-,-[]","[0]-,-[s]-,-[label]-,-[PMI]","[0]-,-[n]-,-[OriginalID]-,-[-2147480000]","[0]-,-[n]-,-[OriginalTypeID]-,-[-2147480000]","[0]-,-[s]-,-[NodePath]-,-[]"},"aux_attributes":{"[4]-,-[IGDAUXID]-,-[4]","[3]-,-[IGDAUXID]-,-[3]","[2]-,-[IGDAUXID]-,-[2]","[1]-,-[IGDAUXID]-,-[1]","[0]-,-[IGDAUXID]-,-[0]"}', true)
 * 
 * */
	v_cache_uri varchar[] ;
	v_uuid 	varchar;
	v_cache_id0 integer;
	v_cache_id1 integer;

	v_new_cache_id		integer;
	v_cur_cache_id		integer;

	v_tables varchar[] := array[ 
								'node', 'aux_node', 'aux_node_to_aux_node',
								'shape' , 'face' , 'edge',
								'shape_instance' , 'face_instance', 'edge_instance',
								'edge_instance_to_aux_node' , 'face_instance_to_aux_node',								
								'aux_attributes' ,
								 'metadata', 'tag',   
								 'aux_value',								
								'index_cache', 'data_stats', 'index_properties', 'index_source', 'index_stats',
								'max_row_count' , 'pending_shapes', 'pending_nodes', 'pending_aux_nodes'  --flags
							]::varchar[];
	v_data varchar[];
	v_tables_data varchar[];
--	v_tables_positions varchar[];

	v_responses varchar[] := array[]::varchar[];


	v_response varchar;
	v_table_data varchar;
	v_table_name varchar;

	v_pending_data boolean := false;
	
	v_exists_tmp_insert_cache bool;

	v_json_data varchar := substring( p_data , 1, length( p_data ) - position( '"' in reverse( p_data ) ) ) ;
	v_json_data_length integer := length( v_json_data );
--	v_t0 timestamp :=  clock_timestamp();
begin	
 
	select array_agg( 
						case when table_data_position != 0 then						
							table_name || '=>' || 
								case 
									when table_name = 'max_row_count' then 
										substring( table_data, '([0-9]+)' )
									when table_name in( 'pending_nodes', 'pending_shapes', 'pending_aux_nodes' )  then								
										substring( table_data, '(t|f)' )
									else
										substring( table_data, table_name || '[\\]?"[\s]*:[\s]*[[][\s]*(.*)[\s]*[]]') 
								end						
						end

					) --, array_agg( table_name || '=>' || table_data_position) 
					
	into  v_tables_data --, v_tables_positions
	from
	(				
		select  table_name, table_data_position, substr( v_json_data, table_data_position, table_data_count ) as table_data
		from
		(	
			select table_name , table_data_position , ( select coalesce( min( p ) - table_data_position + 1, v_json_data_length ) from unnest( positions ) p where p > table_data_position ) as table_data_count 
			from
			(
				select table_name , table_data_position , array_agg( table_data_position ) over() as positions
				from
				(
				--	select table_name , position( substring( v_json_data, '"' || table_name || '[\\]?"[\s]*:' ) in v_json_data ) as table_data_position
					select table_name , position( ( '"' || table_name || '":[' ) in v_json_data ) as table_data_position
					from 
					(
						select t as table_name from unnest(v_tables) t
					) a			
				) a
				where table_data_position != 0
			) a
		) a
	
	)  a
	;
--	raise warning 'import cache %. parsing. time %', p_uri,  (clock_timestamp() - v_t0 ) ;
	-- =============================================================================================================			
--	v_cache_uri := regexp_matches(  p_data , '"?uri"?[\s]*:[\s]"([a-fA-F0-9]{16}-[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4})/([0-9]+)/([0-9]+)"');	
	v_cache_uri := regexp_matches( p_uri, '/([a-fA-F0-9-]{35,60})/([0-9]+)/([0-9]+)/' );


	v_uuid := v_cache_uri[1]::varchar;
	if v_uuid is null then
		v_uuid := substring( p_uri , '([a-fA-F0-9]{40})');
				
		v_cache_id0 := 0;
		v_cache_id1 := 0;
	else 
		v_cache_id0 := v_cache_uri[2]::integer;
		v_cache_id1 := v_cache_uri[3]::integer;	
	end if;		

	-- inserting a cache only done when exists end record on l3d.ins_stats ( 2nd time ) or cache doesn't exists
		
--	raise warning 'v_cache_uri %, v_uuid % , v_cache_id0 %, v_cache_id1 %, v_exists_tmp_insert_cache %'	, v_cache_uri,v_uuid, v_cache_id0, v_cache_id1, v_exists_tmp_insert_cache;

	select max(id) into v_cur_cache_id from l3d."cache" c where c.uuid = v_uuid and c.ext_cache_id0 = v_cache_id0 and c.ext_cache_id1 = v_cache_id1;

	if v_cur_cache_id is null or exists( select 1 from l3d.ins_stats is2 where cache_id = v_cur_cache_id and table_name = 'end' ) then

	 	DISCARD temp;
	 				
		select * into v_new_cache_id from l3d.sp_insert_cache(v_uuid, v_cache_id0, v_cache_id1) ;
	
	
		begin
			
			-- index_cache
			select regexp_matches( d , 
										'"([[:print:]]+)"' 
									||	'[\s]*,[\s]*"([[:print:]]*)"' 
									||	'[\s]*,[\s]*"([[:print:]]*)"' 
									||	'[\s]*,[\s]*"?([0-9]*)"?' 
									||	'[\s]*,[\s]*"([[:print:]]*)"'								
									||	'[\s]*,[\s]*"?([0-9]*)"?'
									||	'[\s]*,[\s]*({.*})'
									) 						
			into v_data 
			from
			(
				select  substr( d , position( '=>' in d ) + 2 ) d  from unnest( v_tables_data ) d where d like 'index_cache%'
			) a;								
			if v_cur_cache_id is null or not exists( select * from l3d.index_cache t where t.cache_id = v_new_cache_id ) then
				insert into l3d.index_cache (url, state, transcoder_group, bytes, cache_id, l3dgen_version, data_layout_version, original_request)
				select v_data[1], v_data[2], v_data[3], v_data[4]::integer, v_new_cache_id, v_data[5], v_data[6]::integer, v_data[7]::json;				
			else 
				update l3d.index_cache u set url = v_data[1], state = v_data[2], transcoder_group = v_data[3], bytes = v_data[4]::integer
										  , l3dgen_version = v_data[5], data_layout_version = v_data[6]::integer, original_request = v_data[7]::json									 
				where u.cache_id = v_new_cache_id ;			
			end if;
		

		
			-- index_stats
			select regexp_matches( d , 
										'"([[:print:]]+)"' 
								||	'[\s]*,[\s]*"?([0-9]*)"?'			
								||	'[\s]*,[\s]*"([[:print:]]*)"'
								||	'[\s]*,[\s]*"([[:print:]]*)"'
								||	'[\s]*,[\s]*"([[:print:]]*)"'
								||	'[\s]*,[\s]*"([[:print:]]*)"'								
									) 						
			into v_data 
			from
			(
				select  substr( d , position( '=>' in d ) + 2 ) d  from unnest( v_tables_data ) d where d like 'index_stats%'
			) a;
		
			if v_cur_cache_id is null or not exists( select * from l3d.index_stats where cache_id = v_new_cache_id ) then		
				insert into l3d.index_stats(queued_date, num_reads, last_modified_date, last_read_date, processing_start_date, processing_end_date, cache_id)
				select v_data[1]::timestamp, v_data[2]::integer, v_data[3]::timestamp, v_data[4]::timestamp, v_data[5]::timestamp, v_data[6]::timestamp, v_new_cache_id ;			
			else		
				update l3d.index_stats set queued_date = v_data[1]::timestamp, num_reads =  v_data[2]::integer, last_modified_date = v_data[3]::timestamp
												, last_read_date = v_data[4]::timestamp, processing_start_date = v_data[5]::timestamp, processing_end_date = v_data[6]::timestamp
				where cache_id = v_new_cache_id ;		
			end if;
	
			-- index_source	
			select regexp_split_to_array( trim( substr( d , position( '=>' in d ) + 2 ), '"' ) , '"[\s]*,[\s]*"' ) d into v_data from unnest( v_tables_data ) d where d like 'index_source%';		
		
			if v_cur_cache_id is null or not exists( select * from l3d.index_source where cache_id = v_new_cache_id ) then		
				insert into l3d.index_source(uri, url, auth_url, mime_type, property, e_tag, cache_id)
				select v_data[1], v_data[2], v_data[3], string_to_array(v_data[4], ';')::varchar[], v_data[5], v_data[6], v_new_cache_id;					
			else		
				update l3d.index_source set uri = v_data[1], url = v_data[2], auth_url = v_data[3], mime_type = string_to_array(v_data[4], ';')::varchar[], property = v_data[5], e_tag = v_data[6]										
				where cache_id = v_new_cache_id;				
			end if;
			
			--index_properties
--			 select array[ substr( d , position( '=>' in d ) + 2 ) ]::varchar[] into v_data from unnest( v_tables_data ) d where d like 'index_properties%';			
			select array[ trim( substr( replace(d, '\"', '"') , position( '=>' in d ) + 2 ) , '"' ) ]::varchar[] into v_data
			from unnest( v_tables_data ) d where d like 'index_properties%' ; 

			
		--	raise warning 'v_data index_properties =>%' , v_data[1];
		
			if v_cur_cache_id is null or not exists( select * from l3d.index_properties where cache_id = v_new_cache_id ) then
				insert into l3d.index_properties(properties, cache_id) select  v_data[1]::json, v_new_cache_id;					
			else
				update l3d.index_properties set properties = v_data[1]::json			
				where cache_id = v_new_cache_id;			
			end if;
		
			--data_stats
			select array[ trim( substr( replace(d, '\"', '"') , position( '=>' in d ) + 2 ) , '"' ) ]::varchar[] into v_data
			from unnest( v_tables_data ) d where d like 'data_stats%' ; 
						
--			select substring( d , '=>[\s]*') from unnest( v_tables_data ) d where d like 'data_stats%'; 			

			--		raise warning 'v_data data_stats =>%' , v_data[1];
		
			if v_cur_cache_id is null or not exists( select * from l3d.data_stats where cache_id = v_new_cache_id ) then
				insert into l3d.data_stats(cache_id , stats) select v_new_cache_id, ( '[' || v_data[1] || ']' )::json	;		
			else 
				update l3d.data_stats set stats = ( '[' || v_data[1] || ']' )::json where cache_id = v_new_cache_id;
			end if;		
		      		
		exception when others then
			declare
				v_err_txt varchar;				
				v_err_ctx text;
			begin
				GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_err_ctx = pg_exception_context; 
				raise exception 'Error inserting cache data message => %, context => % , data => %' , v_err_txt ,  v_err_ctx, v_data;
			end ;
		end; 
		 							
	end if;	


	-- =============================================================================================================
	if v_new_cache_id is not null then
		perform l3d.sp_insert_fetch_cache( v_new_cache_id ) ;		
	elsif not l3d.sp_exists_temporary_table( 'tmp_insert_cache' ) then
		perform l3d.sp_insert_fetch_cache( v_cur_cache_id ) ;
	elsif not exists( select * from tmp_insert_cache c where id = v_cur_cache_id)  then		 
		perform l3d.sp_insert_fetch_cache( v_cur_cache_id ) ;			 						
	end if;
									
	for i in 1..array_length( v_tables_data, 1) loop
	
		select n ,  d	
		into v_table_name, v_table_data
		from 
		( 
			select substr(d , 1 , p - 1) as n, trim( trim( substr( d, p + 2) , '"' ) , '\"' ) as d
			from
			(
				select  position('=>' in d) p, v_tables_data[i] d  from ( select  v_tables_data[i] d ) a									
			) a													
		) a;		
			
	
		if v_table_name = 'max_row_count' then
			v_response := '"' || v_table_name || '":' || v_table_data ;
		elsif v_table_name in ( 'pending_nodes', 'pending_shapes', 'pending_aux_nodes' )  then
			v_pending_data := v_pending_data || v_table_data ~ 't' ;									
			v_response := '"' || v_table_name || '":"' || v_table_data  || '"';
		else 
		
			select 
				coalesce(
					'"' || n || '":{"lastExtId":' ||				
						case
							when n = 'node' then
								( select l3d.sp_insert_nodes( d ) )								
							when n = 'metadata' then							
								( select l3d.sp_insert_metadata( d ) )
							when n = 'tag' then							
								( select l3d.sp_insert_tags( d ) )
							when n = 'aux_node' then							
								( select l3d.sp_insert_aux_nodes( d ) )	
							when n = 'aux_node_to_aux_node' then 
								( select l3d.sp_insert_aux_nodes_relations( d ) )
							when n = 'aux_value' then							
								( select l3d.sp_insert_aux_nodes_values( d ) )																
							when n = 'aux_attributes' then							
								( select l3d.sp_insert_aux_nodes_attributes( d ) )																
							when n = 'shape' then							
								( select l3d.sp_insert_shapes( d ) )																
							when n = 'shape_instance' then							
								( select l3d.sp_insert_shape_instances( d ) )																
							when n in ( 'face_instance', 'edge_instance' ) then							
								( select l3d.sp_insert_face_edge_instances( d ) )																
							when n in ( 'face_instance_to_aux_node', 'edge_instance_to_aux_node' ) then							
								( select l3d.sp_insert_aux_node_topo_relations( d ) )								
							else 
								'0'								
						end ||  '}'
				, '' )::varchar
			into v_response 
			from ( select v_table_name as n, regexp_split_to_array( v_table_data , '"[\s]*,[\s]*"' ) as d) a
			;							
		
		
		end if;
		
--					raise warning 'reponse: % % for %. time response: %', i ,  v_response, v_table_name, ( clock_timestamp() - start_time );
			v_responses := v_responses || v_response;
	
	end loop;	

	if not v_pending_data then
		
		perform  l3d.sp_insert_flush() ;											
		insert into l3d.ins_stats(cache_id, table_name, time_start, records_count, records_commited) select max(id), 'end', clock_timestamp() ,0, 0 from tmp_insert_cache;			
		truncate table tmp_insert_cache;
		return query select null::varchar;
	else 		
		perform  l3d.sp_insert_flush(false) ;	
		return query select  ( '{' || array_to_string( v_responses , ',' ) || '}' )::varchar;			
	end if;	
			
	-- =============================================================================================================
--	raise warning 'import cache %. finished. time %', p_uri,  (clock_timestamp() - v_t0 ) ;
end ;
$function$;
alter function  l3d.import_cache(varchar, varchar, boolean)  OWNER TO i3dhub;

